//
// SMAC.cc       : SMAC/MoteNIC Radio Support for Diffusion
// Authors       : Fabio Silva
//
// Copyright (C) 2000-2002 by the University of Southern California
// $Id: SMAC.cc,v 1.3 2003/02/05 23:14:21 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <stdio.h>
#include <stdlib.h>

#include "SMAC.hh"

SMAC::SMAC() : DiffusionIO()
{
  int fd1;

  // Open MoteNIC'sTOS interface 
  if ((fd1 = open(SMAC_MOTENIC_DEVICE, O_RDWR)) < 0){
    DiffPrint(DEBUG_ALWAYS, "Cannot open " SMAC_MOTENIC_DEVICE " !\n");
    exit(-1);
  }

  DiffPrint(DEBUG_IMPORTANT, "SMAC/MoteNIC: Radio fd: %d\n", fd1);

  in_fds_.push_back(fd1);
  out_fds_.push_back(fd1);

  max_in_descriptor_ = fd1;
}

DiffPacket SMAC::recvPacket(int fd)
{
  short radio_id;
  int diffusion_packet_size;
  int32_t diffusion_id;
  u_int16_t data_len_packet;
  u_int16_t data_len;
  int rf_packet[((MAX_PKT / sizeof(int)) + 1)];
  DiffPacket diffusion_packet = NULL;
  char *rf_payload;
  int rf_packet_size;
  struct hdr_diff *dfh;
  struct AppHeader *rf_header;

  rf_packet_size = read(fd, rf_packet, MAX_PKT);

  DiffPrint(DEBUG_LOTS_DETAILS, "SMAC: Read %d bytes\n", rf_packet_size);

  // Read the entire Sensoria Packet
  rf_header = (struct AppHeader *) &rf_packet[0];
  rf_payload = (char *) ((char *) rf_header + sizeof(struct AppHeader));

  DiffPrint(DEBUG_LOTS_DETAILS,
	    "SMAC: Packet type is %d\n", rf_header->appType);

  // Check if packet is a diffusion packet
  if (rf_header->appType != DIFFUSION_PACKET_TYPE){
    // Print message
    DiffPrint(DEBUG_IMPORTANT,
	      "Received a packet type = %d, discarding...\n",
	      rf_header->appType);

    return NULL;
  }

  radio_id = rf_header->macHdr.fromAddr;
  dfh = HDR_DIFF(rf_payload);

  memcpy(&data_len_packet, &(DATA_LEN(dfh)), sizeof(u_int16_t));
  data_len = ntohs(data_len_packet);

  // Allocate memory for the diffusion packet
  diffusion_packet_size = data_len + sizeof(struct hdr_diff);
  DiffPrint(DEBUG_LOTS_DETAILS, "SMAC: Data len = %d, Total len = %d\n",
	    data_len, diffusion_packet_size);

  diffusion_packet = new int [(diffusion_packet_size / sizeof(int)) + 1];

  if (diffusion_packet == NULL){
    DiffPrint(DEBUG_ALWAYS, "Cannot allocate memory for incoming packet !\n");
    exit(-1);
  }

  // Copy diffusion packet out of the sensoria packet
  memcpy(diffusion_packet, rf_payload, diffusion_packet_size);

  // Extract Sensoria's radio ID and diffusion's ID from packet
  diffusion_id = ntohl(LAST_HOP(dfh));

  // Add radio Id to Cache
  addCache(diffusion_id, radio_id);

  DiffPrint(DEBUG_LOTS_DETAILS,
	    "Received %d bytes from node %d !\n",
	    diffusion_packet_size, diffusion_id);

  return diffusion_packet;
}

void SMAC::sendPacket(DiffPacket pkt, int len, int dst)
{
  short radio_id;
  bool status;
  DiffPacket outgoing_packet = NULL;
  list<int>::iterator itr;
  int retval;

  if ((len + sizeof(struct AppHeader)) > MAX_PKT){
    DiffPrint(DEBUG_ALWAYS,
	      "Error: Packet too big ! Size = %d !\n",
	      (len + sizeof(struct AppHeader)));
    return;
  }

  if (dst != BROADCAST_ADDR){
    status = findInCache(dst, &radio_id);
    if (!status){
      DiffPrint(DEBUG_ALWAYS,
		"Warning: Destination %d not in cache !\n", dst);
      return;
    }

    outgoing_packet = assemblePacket(radio_id, len, pkt);
    for (itr = out_fds_.begin(); itr != out_fds_.end(); ++itr){
      retval = write(*itr, outgoing_packet, MAX_PKT);

      if (retval == MAX_PKT){
	DiffPrint(DEBUG_LOTS_DETAILS,
		  "Sending %d bytes to node %d !\n", len, dst);
      }
      else{
	DiffPrint(DEBUG_IMPORTANT,
		  "Smac driver: write returned %d !\n", retval);
      }
    }
  }
  else{
    // We have a broadcast message
    radio_id = SMAC_BROADCAST;

    outgoing_packet = assemblePacket(radio_id, len, pkt);
    for (itr = out_fds_.begin(); itr != out_fds_.end(); ++itr){
      retval = write(*itr, outgoing_packet, MAX_PKT);

      if (retval == MAX_PKT){
	DiffPrint(DEBUG_LOTS_DETAILS,
	      "Sending %d bytes to broadcast address !\n", len);
      }
      else{
	DiffPrint(DEBUG_IMPORTANT,
		  "Smac driver: write returned %d !\n", retval);
      }
    }
  }
  delete [] outgoing_packet;
}

void SMAC::addCache(int32_t diffusion_id, short radio_id)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache_.begin(); itr != RfCache_.end(); ++itr){
    entry = *itr;
    if (radio_id == entry->radio_id_){
      if (diffusion_id == entry->diffusion_id_){
	return;
      }
      // Something changed for this WINS address
      entry->diffusion_id_ = diffusion_id;
      return;
    }
  }

  // Create a new Cache entry
  DiffPrint(DEBUG_SOME_DETAILS,
	    "SMAC: Adding node %d --> %d to the cache !\n",
	    diffusion_id, radio_id);
  entry = new RFIdCacheEntry(diffusion_id, radio_id);
  RfCache_.push_back(entry);
}

bool SMAC::findInCache(int32_t diffusion_id, short *radio_id)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache_.begin(); itr != RfCache_.end(); ++itr){
    entry = *itr;

    if (diffusion_id == entry->diffusion_id_){
      // Entry found !
      *radio_id = entry->radio_id_;
      return true;
    }
  }

  // Entry not in the cache
  return false;
}

DiffPacket SMAC::assemblePacket(short radio_id, int len, DiffPacket pkt)
{
  DiffPacket outgoing_packet = NULL;
  char *packet_payload;
  struct AppHeader *rf_header;

  // Allocate memory for packet + header
  outgoing_packet = new int[MAX_PKT / sizeof(int)];
  rf_header = (struct AppHeader *) outgoing_packet;
  packet_payload = (char *) rf_header + sizeof(struct AppHeader);

  // Fill rf header
  rf_header->macHdr.toAddr = radio_id;
  rf_header->appType = (uint8_t) DIFFUSION_PACKET_TYPE;

  // Copy packet
  memcpy(packet_payload, pkt, len);

  return outgoing_packet;
}
