//
// emstar.cc     : Emstar Device Class Support for Diffusion
// Authors       : Fabio Silva
//
// Copyright (C) 2000-2003 by the University of Southern California
// $Id: emstar.cc,v 1.4 2003/06/06 00:54:08 fabio Exp $
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License,
// version 2, as published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
//

#include <stdio.h>
#include <stdlib.h>

#include "emstar.hh"

Emstar::Emstar(int32_t node_id, int32_t group_id,
	       bool under_emsim) : DiffusionIO()
{
  char device_name[100];
  pd_filter_t rf_filter;
  int status;
  int fd1;

  // Figure out which device we should use
  if (!under_emsim){
    // Standard device for emstar-class devices
    sprintf(device_name, EMSTAR_DEVICE);
  }
  else{
    // When running under emsim, we have to use group and node ids to
    // figure out which device we should use
    sprintf(device_name,
	    EMSTAR_DEVICE_PREFIX "/group%d/node%d" EMSTAR_DEVICE_SUFFIX,
	    group_id, node_id);
  }

  // Open emstar device
  if ((fd1 = open(device_name, O_RDWR | O_NONBLOCK)) < 0){
    DiffPrint(DEBUG_ALWAYS, "Cannot open %s !\n", &device_name[0]);
    exit(-1);
  }

  DiffPrint(DEBUG_IMPORTANT, "Emstar-class device opened, fd: %d\n", fd1);

  in_fds_.push_back(fd1);
  out_fds_.push_back(fd1);

  max_in_descriptor_ = fd1;
  
  // Add filters to the radio devices so we just receive our packets
  rf_filter.data[0] = PKT_TYPE_DIFFUSION;
  rf_filter.len = 1;

  status = ioctl(fd1, PD_SET_FILTER, &rf_filter);
  if (status < 0){
    DiffPrint(DEBUG_ALWAYS,
	      "Cannot set filter for radio %s !\n", &device_name[0]);
    exit(-1);
  }
}

DiffPacket Emstar::recvPacket(int fd)
{
  struct link_pkt *rf_pkt_hdr;
  link_pkt_type_t rf_packet_type;
  if_id_t radio_id;
  int diffusion_packet_size;
  int32_t diffusion_id;
  DiffPacket rf_packet = NULL;
  DiffPacket diffusion_packet = NULL;
  int flag;
  char *rf_payload;
  int rf_packet_size;
  struct hdr_diff *dfh;

  // Determine packet size
  flag = ioctl(fd, PD_GET_FRONT_SIZE, &rf_packet_size);
  if (flag != 0){
    DiffPrint(DEBUG_ALWAYS, "Cannot determine packet size ! Error Code = %d\n",
	      flag);
    exit(-1);
  }

  // Convert from network to host order
  rf_packet_size = ntohl(rf_packet_size);

  if (rf_packet_size <= 0){
    DiffPrint(DEBUG_ALWAYS, "Incoming packet has size <= 0 !\n");
    return NULL;
  }

  // Allocate memory for the incoming packet
  rf_packet = new int [(rf_packet_size / sizeof(int)) + 1];

  if (rf_packet == NULL){
    DiffPrint(DEBUG_ALWAYS,
	      "Cannot allocate memory for the incoming radio packet !\n");
    exit(-1);
  }

  // Read the entire Emstar Packet
  read(fd, rf_packet, rf_packet_size);
  rf_pkt_hdr = (struct link_pkt *) rf_packet;
  rf_payload = (char *) rf_pkt_hdr->data;

  // Check if packet is a diffusion packet
  rf_packet_type = rf_pkt_hdr->type;
  if (rf_packet_type != PKT_TYPE_DIFFUSION){
    // Print message
    DiffPrint(DEBUG_IMPORTANT,
	      "Received a packet type = %d, discarding...\n",
	      rf_packet_type);

    delete [] rf_packet;
    return NULL;
  }

  // Allocate memory for the diffusion packet
  diffusion_packet_size = rf_packet_size - sizeof(struct link_pkt);
  diffusion_packet = new int [(diffusion_packet_size / sizeof(int)) + 1];

  if (diffusion_packet == NULL){
    DiffPrint(DEBUG_ALWAYS, "Cannot allocate memory for incoming packet !\n");
    exit(-1);
  }

  // Copy diffusion packet out of the emstar packet
  memcpy(diffusion_packet, rf_payload, diffusion_packet_size);

  // Extract Emstar's radio ID and diffusion's ID from packet
  dfh = HDR_DIFF(diffusion_packet);
  diffusion_id = ntohl(LAST_HOP(dfh));
  radio_id = rf_pkt_hdr->src.id;

  // Add rf id to cache
  addCache(diffusion_id, radio_id);

  // We don't need the emstar packet anymore
  delete rf_packet;

  DiffPrint(DEBUG_LOTS_DETAILS,
	    "Received %d bytes from node %d !\n",
	    diffusion_packet_size, diffusion_id);

  return diffusion_packet;
}

void Emstar::sendPacket(DiffPacket pkt, int len, int dst)
{
  DiffPacket outgoing_packet = NULL;
  list<int>::iterator itr;
  if_id_t radio_id;
  int status;

  if (dst != BROADCAST_ADDR){
    if (!findInCache(dst, &radio_id)){
      DiffPrint(DEBUG_ALWAYS,
		"Warning: Destination %d not in cache !\n", dst);
      return;
    }

    DiffPrint(DEBUG_LOTS_DETAILS, "Sending %d bytes to node %d !\n", len, dst);
  }
  else{
    // We have a broadcast message
    radio_id = LINK_BROADCAST;

    DiffPrint(DEBUG_LOTS_DETAILS,
	      "Sending %d bytes to broadcast address !\n", len);
  }

  outgoing_packet = assemblePacket(radio_id, len, pkt);
  for (itr = out_fds_.begin(); itr != out_fds_.end(); ++itr){
    status = write(*itr, outgoing_packet, (len + sizeof(struct link_pkt)));
    if (status != (ssize_t) (len + sizeof(struct link_pkt)))
      DiffPrint(DEBUG_ALWAYS, "Emstar: Write returned %d !\n", status);
  }

  delete [] outgoing_packet;
}

void Emstar::addCache(int32_t diffusion_id, if_id_t radio_id)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache_.begin(); itr != RfCache_.end(); ++itr){
    entry = *itr;
    if (radio_id == entry->radio_id_){

      // Do nothing if Diffusion ID remains the same
      if (diffusion_id == entry->diffusion_id_){
	return;
      }

      // Diffusion ID changed for this rf address
      entry->diffusion_id_ = diffusion_id;
      return;
    }
  }

  // Create a new Cache entry
  entry = new RFIdCacheEntry(diffusion_id, radio_id);
  RfCache_.push_back(entry);
}

bool Emstar::findInCache(int32_t diffusion_id, if_id_t *radio_id)
{
  RFIdCache::iterator itr;
  RFIdCacheEntry *entry;

  for (itr = RfCache_.begin(); itr != RfCache_.end(); ++itr){
    entry = *itr;

    if (diffusion_id == entry->diffusion_id_){
      // Entry found !
      *radio_id = entry->radio_id_;
      return true;
    }
  }

  // Entry not in the cache
  return false;
}

DiffPacket Emstar::assemblePacket(if_id_t radio_id, int len, DiffPacket pkt)
{
  DiffPacket outgoing_packet = NULL;
  char *packet_payload;
  struct link_pkt *rf_pkt_hdr;

  // Allocate memory for packet + header
  outgoing_packet = new int[((len + sizeof(struct link_pkt)) / sizeof(int))];
  rf_pkt_hdr = (struct link_pkt *) outgoing_packet;
  packet_payload = (char *) rf_pkt_hdr->data;

  // Fill rf header
  rf_pkt_hdr->dst.id = radio_id;
  rf_pkt_hdr->type = PKT_TYPE_DIFFUSION;

  // Copy packet
  memcpy(packet_payload, pkt, len);

  return outgoing_packet;
}
